#ifndef RTOS_H
#define RTOS_H

/* SPM - below SPM_LIMIT */
#define SPM_LIMIT           0x00001000

/* Code and user data area: SPM_LIMIT to 0x80000000 */
#define SWITCH_STACK        0x20000000
#define SYSTEM_STACK        0x3f000000
#define EMPTY_SPACE_AREA    0x50000000
#define USER_LIMIT          0x80000000

/* Devices area */
#define TIMER_REGISTER      0x80010000
#define BASE_REGISTER_AREA  0x80020000

/* CPU flags */
#ifndef MSR_IE
#define MSR_IE              0x002
#endif
#ifndef MSR_BE
#define MSR_BE              0x100
#endif
#define MSR_MASK            (MSR_IE | MSR_BE)
#define MODE_USER           (MSR_IE | MSR_BE)
#define MODE_SYSCALL        (0)
#define MODE_SYSTEM         (MSR_BE)


#define BLOCK_SIZE_SHIFT    7
#define NUM_BLOCKS_SHIFT    4
#define MAX_THREADS         16

#define EXT_MEM_LATENCY     49  
#define EXT_MEM_BUS_WIDTH   4
#define EXT_MEM_TRANS_BYTES 64

#define TIMER_QUANTUM       1000

#define NUM_BLOCKS          (1 << NUM_BLOCKS_SHIFT)
#define BLOCK_SIZE          (1 << BLOCK_SIZE_SHIFT)

#define ICACHE_STORE_SIZE   (BLOCK_SIZE * NUM_BLOCKS / 2)
#define DCACHE_STORE_SIZE   (BLOCK_SIZE * NUM_BLOCKS / 2)
#define ICACHE_LINE_SIZE    EXT_MEM_TRANS_BYTES
#define DCACHE_LINE_SIZE    EXT_MEM_TRANS_BYTES
/*#define UNLIMITED_TRANSFERS*/

#define OPEN_DISCARD_FLAG   0x4000
#define OPEN_SCRATCH_FLAG   0x2000
#define OPEN_DEFER_FLAG     0x1000
#define OPEN_SIZE_BITS      0x00ff

#define AKD __attribute__ ((section (".kernelstuff.d")))
#define AKT __attribute__ ((section (".kernelstuff.t"))) 
#define ALIGN __attribute__ ((aligned(BLOCK_SIZE)))


#define TRC_CAPT_CALL_FUNCTION      1
#define TRC_CAPT_RETURN_FUNCTION    2
#define TRC_IN_RTI                  4
#define TRC_TASK_START              5
#define TRC_TASK_END                6

#ifndef FROM_ASM

/* defined in asm */
extern void * call_function(void * fn, unsigned block_info_word, ...);
extern void * memcpy(void * dest, const void * src, unsigned size);
extern void * memmove(void * dest, const void * src, unsigned size);
extern void _open(void * base, unsigned block_number, unsigned num_blocks);
extern void print(const char * text);
extern void * scheduler(void);
extern void exit(int rc);
extern void _exit(int rc);
extern void notify_task_name(const char * name, void * function);
extern volatile unsigned current_time;
extern volatile unsigned char __end_spm;

/* defined here */
#define _close(block_number, num_blocks) \
    _open((void *) 0xffffffff, (block_number), (num_blocks))
#define open(handle, base, size) \
    do { const unsigned num_blocks = ((size) + BLOCK_SIZE - 1) / BLOCK_SIZE; \
    _open((void *) (base), handle, (num_blocks)); \
    handle += num_blocks; } while(0)
#define close(handle, base, size) \
    do { const unsigned num_blocks = ((size) + BLOCK_SIZE - 1) / BLOCK_SIZE; \
    handle -= num_blocks; _close(handle, (num_blocks)); } while(0)

/* defined in C */
void * scheduler(void) AKT;
void register_task(const char * name, void * function, 
                unsigned block_info_word,
                unsigned offset, unsigned period) AKT;
void kernel_start(void) AKT;
void crash(void) AKT;

#endif
#endif

