
import random, os, math

# Run the program once to find these:
CS1 = 401       # printed by the program
CS2 = 387       # printed by the program
B = 4321        # the largest number at the end of any line in searcher.log



NUM_RUNS = 1000
SZ = 15000000
S_DIR = "../searcher"
TASK_LIST = ["fir", "matmult", "cnt", "bs",
    "bsort100", "compress", "crc",
    "duff", "expint", "fdct", "insertsort",
    "janne_complex", "jfdctint" ]

FIELDSEP = 15
class Task_Data: 
    def __init__(self):
        self.completes = []
        self.bcets = []
        self.wcets = []
        self.acets = []
        self.preempts = []
        self.wcet = 0
        self.taskset = []


def run():
    task_list = TASK_LIST[:]
    task_wcet_data = dict()

    rc = os.system("make -C ../simulator")
    assert rc == 0
    rc = os.system("rm -rf %s" % S_DIR)
    os.mkdir(S_DIR)

    all_report = file("searcher.log", "wt")

    missing = []
    completion = NUM_RUNS
    task_set = -1
    while completion > 0:
        task_set += 1
        all_report.flush()
        all_report.write("%u " % task_set)
        r = random.Random(hash((task_set, "poppy")))

        if task_set != 0:
            # normal phase
            task_list.sort()
            r.shuffle(task_list)
        else:
            # WCET capture
            print 'Generating WCET capture schedule'
            task_list.sort()
            task_list.reverse()

        period = SZ
        offset = 0
        C = []
        T = []
        R = []
        out = file("tasks.lst", "wt")
        for task in task_list:

            if task_set != 0:
                # normal phase
                wcet = float(task_wcet_data[task[:8]])
                min_period = int(wcet * 2.0)
                assert min_period < (SZ / 2), (task, min_period)
                max_period = max(SZ / 4, min_period * 2)
                assert max_period < (SZ / 2)
                period = r.randint(min_period, max_period)
                offset = r.randint(0, period)
                C.append(wcet)
                T.append(float(period))
                R.append(wcet)

            out.write("%s %u %u\n" % (task, offset, period))

            if task_set == 0:
                # WCET capture phase
                offset += (SZ / len(task_list))

        out.close()

        # Schedulability analysis, see Burns and Wellings 4th ed page 416
        schedulable = True
        if task_set != 0:
            sr = file("%s/%04u.rep" % (S_DIR, task_set), "wt")
            sr.write("CS1 = %u  CS2 = %u\n" % (CS1, CS2))
            for i in xrange(0, len(task_list), 1):
                new_R = 0
                while (R[i] <= T[i]) and (abs(new_R - R[i]) > 0.1):
                    R[i] = new_R
                    new_R = CS1 + CS2 + C[i] + B
                    for j in xrange(i, len(task_list), 1):
                        new_R += math.ceil(R[i]/T[j]) * (
                                CS1 + CS2 + C[j])
                
                sr.write('%16s %9u %9u %9u' % (task_list[i], C[i], T[i], R[i]))
                if T[i] < R[i]:
                    sr.write(' no\n')
                    schedulable = False
                else:
                    sr.write(' yes\n')

            sr.close()

        if not schedulable:
            all_report.write(" /sched\n")
            continue

        #if task_set != 0:
        #    completion -= 1
        #    all_report.write("\n")
        #    continue

        print 'Recompile', task_set, completion
        os.system("make clean 2>/dev/null >/dev/null")
        rc = os.system("make rtos.elf rtos_cache.elf 2>cap2 >cap1")
        assert rc == 0, "Did not compile"

        file("%s/%04u.lst" % (S_DIR, task_set), "wb").write(
                file("tasks.lst", "rb").read())

        print 'Carousel simulation', task_set
        rc = os.system(
            "../simulator/simulator rtos.elf "
                "/dev/null %u >report.txt" % SZ)
        carousel = basic_parse("report.txt")
        assert rc == 0, rc
        assert carousel != None

        print 'Cache simulation', task_set
        rc = os.system(
            "../simulator/simulator rtos_cache.elf "
                "/dev/null %u >report_cache.txt" % SZ)
        cache = basic_parse("report_cache.txt")
        if cache == None:
            assert ((rc >> 8) & 0xff) == 70, rc # deadline overrun

        print 'Cache (wb) simulation', task_set
        rc = os.system(
            "../simulator/simulator rtos_cache.elf "
                "/dev/null %u w >report_cache_w.txt" % SZ)
        cache_wb = basic_parse("report_cache_w.txt")
        if cache_wb == None:
            assert ((rc >> 8) & 0xff) == 70, rc # deadline overrun

        file("%s/%04u.sel" % (S_DIR, task_set), "wt").write(carousel)

        if cache == None:
            all_report.write(" /che")
        else:
            file("%s/%04u.che" % (S_DIR, task_set), "wt").write(cache)

        if cache_wb == None:
            all_report.write(" /cwb")
        else:
            file("%s/%04u.cwb" % (S_DIR, task_set), "wt").write(cache_wb)

        if task_set == 0:
            # WCET capture phase
            task_data = dict()
            adv_parse(task_data, "%s/%04u.sel" % (S_DIR, task_set), None)
            for ((mode, task_name), td) in task_data.items():
                if task_name.startswith("USR"):
                    task_name = task_name[4:]
                    print '%s ---> %u' % (task_name, td.wcet)
                    assert td.wcet != 0
                    task_wcet_data[task_name] = td.wcet
    
            # As defined by Burns and Wellings
            assert CS1 >= task_data[(None, "SYS sch from")].wcet
            assert CS2 >= task_data[(None, "SYS sch to")].wcet

        else:
            completion -= 1
            for line in file("report.txt"):
                if line.startswith("CS time "):
                    fields = line.split()
                    all_report.write(" %s" % fields[-1])
                    
        all_report.write("\n")
        print ''

    print missing

def basic_parse(in_name):
    active = False
    records = []
    current = None
    for line in file(in_name):
        line = line.rstrip()
        if line.startswith("Stop normally"):
            active = True

        elif active and (line.startswith("USR") or line.startswith("SYS")):
            fields = line.split()
            current = [line]
            records.append((' '.join(fields[:2]), current))

        elif current and (len(line) == 0):
            current = None

        elif current and len(line):
            current.append(line)

    records.sort()
    if len(records) == 0:
        return None

    out = []
    for (key, lines) in records:
        out.append('\n'.join(lines))
        out.append("")

    return '\n'.join(out)

def adv_parse(task_data, file_name, mode):
    records = []
    td = None
    for line in file(file_name):
        line = line.rstrip()
        if (line.startswith("USR") or line.startswith("SYS")):
            task_name = line[:FIELDSEP].strip()
            
            k = (mode, task_name)
            td = task_data.get(k, None)
            if td == None:
                task_data[k] = td = Task_Data()

            fields = line[FIELDSEP:].split()
            td.completes.append(int(fields[1]))
            td.bcets.append(int(fields[-3]))
            td.acets.append(int(fields[-2]))
            td.wcets.append(int(fields[-1]))
            td.taskset.append(file_name)

        elif line.find(" syscs ") > 0:
            fields = line[FIELDSEP:].split()
            td.bcets[-1] += (int(fields[-3]))
            td.acets[-1] += (int(fields[-2]))
            td.wcets[-1] += (int(fields[-1]))
            td.wcet = max(td.wcets)
    
        elif line.find(" preempts ") > 0:
            fields = line[FIELDSEP:].split()
            td.preempts.append(int(fields[-1]))


if __name__ == "__main__":
    run()

